import { Body, Controller, Get, Param, Post, UseFilters } from "@nestjs/common";
import { AllExceptionsHandler } from "../exception.handler";
import { AttestationManagerClient } from "@ocm-engine/clients";
import {
  CreateCredentialDefinitionRequsetDto,
  CreateSchemaRequestDto,
  CRED_DEF_CREATE,
  CRED_ISSUE,
  CRED_LIST,
  CRED_OFFER_ACCEPT,
  CRED_OFFER_LIST,
  GatewayAcceptedResponseDto,
  IssueCredentialRequestDto,
  MakeBasicMessageRequestDto,
  MESSAGE_MAKE,
  SCHEMA_CREATE,
  SCHEMA_GET,
  SCHEMA_LIST,
  IdReqDto,
} from "@ocm-engine/dtos";
import {
  ApiBadRequestResponse,
  ApiInternalServerErrorResponse,
  ApiOperation,
  ApiResponse,
} from "@nestjs/swagger";

@UseFilters(AllExceptionsHandler)
@Controller("v1")
export class AttestationController {
  constructor(private readonly amClient: AttestationManagerClient) {}

  @Post("/schemas")
  @ApiResponse({
    status: 201,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: ["name must be a string", "name should not be empty"],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Create schema",
    description:
      "Method will create schema. The id of the response will be matched when you receive event from the websocket",
    tags: ["Schema"],
  })
  async createSchema(@Body() schemaDto: CreateSchemaRequestDto) {
    return this.amClient.sendPayload<CreateSchemaRequestDto>({
      pattern: "schemas",
      payload: {
        source: "/schema",
        data: schemaDto,
        type: SCHEMA_CREATE,
      },
    });
  }

  @Post("/schemas-by-id")
  @ApiResponse({
    status: 201,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "schemaId must be a string",
                "schemaId should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Get schema by id",
    description:
      "Method will fetch specific schema or return null. The id of the response will be matched when you receive event from the websocket",
    tags: ["Schema"],
  })
  async getSchemaById(@Body() schemaRequestDto: IdReqDto) {
    return this.amClient.sendPayload<IdReqDto>({
      pattern: "schemas",
      payload: {
        source: "/schemas-by-id",
        data: schemaRequestDto,
        type: SCHEMA_GET,
      },
    });
  }

  @Get("/schemas")
  @ApiResponse({
    status: 200,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "List all schemas",
    description:
      "Method will fetch all schemas. The id of the response will be matched when you receive event from the websocket",
    tags: ["Schema"],
  })
  async listSchema() {
    return this.amClient.sendPayload<null>({
      pattern: "schemas",
      payload: {
        source: "/schema",
        data: null,
        type: SCHEMA_LIST,
      },
    });
  }

  @Post("credentials/definition")
  @ApiResponse({
    status: 201,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "schemaId must be a string",
                "schemaId should not be empty",
                "tag must be a string",
                "tag should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Create credential definition",
    description:
      "Method create credential definition. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  createCredentialDefinition(
    @Body() credentialDefinitionDto: CreateCredentialDefinitionRequsetDto,
  ) {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "/credential/definition",
        data: credentialDefinitionDto,
        type: CRED_DEF_CREATE,
      },
    });
  }

  @Post("credentials/issue")
  @ApiResponse({
    status: 201,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "connectionId must be a string",
                "connectionId should not be empty",
                "credentialDefinitionId must be a string",
                "credentialDefinitionId should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Issue credential",
    description:
      "Method issue credential, it will create an offer and send it to specified receiver (connectionId). The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  issueCredential(@Body() issueCredentialDto: IssueCredentialRequestDto) {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "credential/issue",
        data: issueCredentialDto,
        type: CRED_ISSUE,
      },
    });
  }

  @Get("/credentials")
  @ApiResponse({
    status: 200,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "List all credential",
    description:
      "Method list credential definition no filters applied. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials"],
  })
  credentials() {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "credentials",
        data: null,
        type: CRED_LIST,
      },
    });
  }

  @Get("/credentials/offers")
  @ApiResponse({
    status: 200,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "List unaccepted credential offers",
    description:
      "Method list offers that are received, but not accepted. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Offers"],
  })
  getCredentialOffers() {
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "credentials/offers",
        data: null,
        type: CRED_OFFER_LIST,
      },
    });
  }

  @Post("/credentials/offers/:credential_record_id/accept")
  @ApiResponse({
    status: 200,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Accept credential offers",
    description:
      "Method list accept credential offer. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Offers"],
  })
  acceptCredential(@Param("credential_record_id") credentialRecordId: string) {
    const data = new IdReqDto();
    data.id = credentialRecordId;
    return this.amClient.sendPayload({
      pattern: "credentials",
      payload: {
        source: "credential/:id/accept",
        data,
        type: CRED_OFFER_ACCEPT,
      },
    });
  }

  @Post("/messages")
  @ApiResponse({
    status: 200,
    description:
      "Request is accepted for execution, the response id will match the event id received from the web socket",
    type: GatewayAcceptedResponseDto,
  })
  @ApiBadRequestResponse({
    description: "Validation error",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 400,
            },

            message: {
              type: "array",
              example: [
                "connectionId must be a string",
                "connectionId should not be empty",
                "message must be a string",
                "message should not be empty",
              ],
            },
            error: {
              type: "string",
              example: "Bad Request",
            },
          },
        },
      },
    },
  })
  @ApiInternalServerErrorResponse({
    description:
      "Error in sending data to attestation manager. This error shows that attestation manager could not convert request to event or attestation manager could not send the event to the broker.",
    content: {
      "application/json": {
        schema: {
          type: "object",
          properties: {
            statusCode: {
              type: "number",
              example: 500,
            },

            message: {
              type: "string",
              example: "connect ECONNREFUSED 0.0.0.0.0:1234",
            },
          },
        },
      },
    },
  })
  @ApiOperation({
    summary: "Send basic message",
    description:
      "Method will send basic message to a connection. The id of the response will be matched when you receive event from the websocket",
    tags: ["Credentials Offers"],
  })
  sendMeesage(@Body() message: MakeBasicMessageRequestDto) {
    return this.amClient.sendPayload({
      pattern: "messages",
      payload: {
        source: "messages",
        data: message,
        type: MESSAGE_MAKE,
      },
    });
  }
}
